# Copyright (C) 2004 Laurent Sansonetti
#
# Tengen is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Tengen is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with Tengen; see the file COPYING.  If not,
# write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

class Tengen::GTP::Vertex
    attr_accessor :x, :y, :stone
    def busy?
        stone != nil
    end
end

module Tengen::UI
    class Board < Gnome::Canvas 
        attr_accessor :decorations_font

        HOSHIS = {
            5 => %w{B2 B4 C3 D2 D4},
            6 => %w{C3 C4 D3 D4},
            7 => %w{C3 C5 D4 E3 E5},
            8 => %w{C3 C6 F3 F6},
            9 => %w{C3 C7 E5 G3 G7},
            10 => %w{C3 C8 H3 H8},
            11 => %w{C3 C9 F6 J3 J9},
            12 => %w{D4 D9 J4 J9},
            13 => %w{D4 D7 D10 G4 G7 G10 K4 K7 K10},
            14 => %w{D4 D11 L4 L11},
            15 => %w{D4 D8 D12 H4 H8 H12 M4 M8 M12},
            16 => %w{D4 D13 N4 N13},
            17 => %w{D4 D9 D14 J4 J9 J14 O4 O9 O14},
            18 => %w{D4 D15 P4 P15},
            19 => %w{D4 D10 D16 K4 K10 K16 Q4 Q10 Q16}
        }

        def initialize(parent, game)
            super(true)
            game.add_observer(self)
    
            @game = game
            @size = game.boardsize
            @parent = parent
            @decorations_font = 'Sans Bold 14'
            @captures = {}
            
            self.width_request = self.height_request = edge
            set_scroll_region(0, 0, edge, edge)
            refresh_vertices
            draw_goban
            
            signal_connect('motion-notify-event') do |widget, event|
                on_mouse_move(event.x, event.y) if should_accept_events? 
            end

            signal_connect('key-press-event') do |widget, event|
                on_keyboard_event(event.keyval) if should_accept_events?
            end
        end
   
        def update(game, player, action)
            if action.pass?
                # nothing to do
            elsif action.undo? or should_redraw?(player)
                redraw_from_engine
            else
                vertex = @vertices.find { |vertex| vertex == action.vertex }
                raise "Invalid move" if vertex.nil?
                pixbuf = case player.color
                    when Tengen::Player::BLACK then Pixmaps::BLACK
                    when Tengen::Player::WHITE then Pixmaps::WHITE
                    else
                        raise "Invalid color"
                end
                vertex.stone = draw_stone(vertex.x, vertex.y, pixbuf)
            end
        end
   
        def freeze
            @frozen = true
            @lastitem.hide if @lastitem
        end
  
        def frozen?
            @frozen
        end
            
        def redraw_from_engine
            Tengen.debug("Redrawing the entire board...")
            black_vertices = @game.black_player.stones
            white_vertices = @game.white_player.stones
            @vertices.each do |vertex|
                if black_vertices.include?(vertex) 
                    unless vertex.busy?
                        Tengen.debug("Adding #{vertex}.")
                        vertex.stone = draw_stone(vertex.x, vertex.y, 
                                                  Pixmaps::BLACK)
                    end
                elsif white_vertices.include?(vertex)
                    unless vertex.busy?
                        Tengen.debug("Adding #{vertex}.")
                        vertex.stone = draw_stone(vertex.x, vertex.y, 
                                                  Pixmaps::WHITE)
                    end
                elsif vertex.busy?
                    Tengen.debug("Removing #{vertex}.")
                    vertex.stone.hide
                    vertex.stone = nil
                end
            end
            Tengen.debug("Redraw complete.")
        end

        #######
        private
        #######
   
        def should_accept_events?
            not frozen? and @game.current_player.human?
        end

        def draw_stone(x, y, pixbuf)
            Gnome::CanvasPixbuf.new(self.root,
                                    { :pixbuf => pixbuf,
                                      :x => x - (pixbuf.width / 2),
                                      :y => y - (pixbuf.height / 2) })
        end
   
        def draw_hoshi(x, y)
            Gnome::CanvasEllipse.new(self.root,
                                     { :x1 => x - 2,
                                       :y1 => y - 2,
                                       :x2 => x + 2,
                                       :y2 => y + 2,
                                       :outline_color => 'black',
                                       :fill_color => 'black' })
        end
   
        BORDER_MARGIN = 50 
        SQUARE_LENGTH = 35 
    
        def edge_without_margin
            (@size - 1) * SQUARE_LENGTH
        end
    
        def edge
            edge_without_margin + (BORDER_MARGIN * 2)
        end
    
        def numbers
            (1..@size).to_a
        end
            
        def letters
            a = ("A".."Z").to_a[0..@size]
            a.delete("I")
            return a
        end
    
        def draw_goban
            Gnome::CanvasRect.new(self.root, 
                                  { :x1 => 0, :y1 => 0,
                                    :x2 => edge, :y2 => edge,
                                    :fill_color => 'orange' })
    
            points = []
            BORDER_MARGIN.step(edge_without_margin + BORDER_MARGIN,
                               SQUARE_LENGTH) { |i| points << i }
  
            first_mid_margin = points.first - (BORDER_MARGIN / 2)
            last_mid_margin = points.last + (BORDER_MARGIN / 2)
            
            points.each_with_index do |p, i|
                # print lines
                Gnome::CanvasLine.new(self.root,
                                      { :points => [[points.first, p],
                                                    [points.last, p]],
                                        :fill_color => 'black' })
                Gnome::CanvasLine.new(self.root,
                                      { :points => [[p, points.first],
                                                    [p, points.last]],
                                        :fill_color => 'black' })
                
                # print numerotations
                Gnome::CanvasText.new(self.root,
                                      { :text => letters[i],
                                        :fill_color => 'black',
                                        :font => @decorations_font,
                                        :x => p,
                                        :y => first_mid_margin })
                Gnome::CanvasText.new(self.root,
                                      { :text => letters[i],
                                        :fill_color => 'black',
                                        :font => @decorations_font,
                                        :x => p,
                                        :y => last_mid_margin })
                Gnome::CanvasText.new(self.root,
                                      { :text => numbers[-i-1].to_s,
                                        :fill_color => 'black',
                                        :font => @decorations_font,
                                        :x => first_mid_margin,
                                        :y => p })
                Gnome::CanvasText.new(self.root,
                                      { :text => numbers[-i-1].to_s,
                                        :fill_color => 'black',
                                        :font => @decorations_font,
                                        :x => last_mid_margin,
                                        :y => p })
            end

            # print hoshis
            HOSHIS[@size].each do |coord|
                vertex = @vertices.find { |x| x == coord }
                draw_hoshi(vertex.x, vertex.y)
            end
        end
    
        def refresh_vertices
            @vertices = [] 
            y = BORDER_MARGIN 
            numbers.reverse.each do |number|
                x = BORDER_MARGIN 
                letters.each do |letter|
                    vertex = Tengen::GTP::Vertex.new("#{letter}#{number}")
                    vertex.x = x
                    vertex.y = y
                    @vertices << vertex
                    x += SQUARE_LENGTH
                end
                y += SQUARE_LENGTH
            end
        end
    
        VERTEX_THR = (SQUARE_LENGTH / 2) 
        def translate_vertex(x, y)
            @vertices.find do |vertex|
                x >= vertex.x - VERTEX_THR and x <= vertex.x + VERTEX_THR and
                y >= vertex.y - VERTEX_THR and y <= vertex.y + VERTEX_THR
            end
        end

        def play_stone(vertex)
            return if vertex.busy? 
            @lastitem.hide 
            begin
                @game.current_player.play(vertex)
            rescue Tengen::GTP::Error => error
                ErrorDialog.new(@parent, error.message, 
                                "Suicide is forbidden, your stone would be " \
                                "dead there.")
            end
        end
        
        def draw_selection_cursor(vertex)
            return if @lastvertex == vertex or vertex.busy?
            alpha_pixbuf = case @game.current_player.color
                when Tengen::Player::BLACK then Pixmaps::BLACK_ALPHA
                when Tengen::Player::WHITE then Pixmaps::WHITE_ALPHA
                else
                    raise "Invalid color"
            end
            @lastitem.hide unless @lastitem.nil?
            @lastitem = draw_stone(vertex.x, vertex.y, alpha_pixbuf)
            @lastitem.signal_connect('event') do |item, event|
                play_stone(vertex) \
                    if event.event_type == Gdk::Event::BUTTON_PRESS
            end
            @lastvertex = vertex
        end

        def on_mouse_move(x, y)
            if vertex = translate_vertex(x, y)
                draw_selection_cursor(vertex)
            end
        end
        
        def on_keyboard_event(keyval)
            current_vertex = (@lastvertex or @vertices[@vertices.length / 2])
            letter = letters.index(current_vertex.letter)
            number = numbers.index(current_vertex.number)
           
            case keyval
                when Gdk::Keyval::GDK_Left
                    letter -= 1 if letter > 0
                   
                when Gdk::Keyval::GDK_Right
                    letter += 1 if letter < @size - 1
                    
                when Gdk::Keyval::GDK_Down
                    number -= 1 if number > 0

                when Gdk::Keyval::GDK_Up
                    number += 1 if number < @size - 1
                    
                when Gdk::Keyval::GDK_space
                    play_stone(current_vertex)
            end

            next_vertex = @vertices.find { |x| x.letter == letters[letter] and 
                                               x.number == numbers[number] }
            
            draw_selection_cursor(next_vertex)
        end
   
        def should_redraw?(player)
            @captures[player] ||= player.captures
            r = player.captures > @captures[player]
            @captures[player] = player.captures
            return r
        end
    end
end
